/*********
  Rui Santos:   Complete project details at https://RandomNerdTutorials.com/esp32-web-server-sent-events-sse/
  Server Sent Events
  Display SPIFFS DIRECTORY https://www.esp8266.com/viewtopic.php?t=20417

  https://forum.arduino.cc/t/esp32-problem-with-asynch-web-server-and-server-sent-events-title-edited/870010/24
*********/

/*
   2.04: added spiffs library and real data handling, plus led config
   version 2.05 was running with leds and sensor connected but still guru errors -
   I changed the loop to a switch case structure to seperate the events and with
   timerDelay=100ms the loop completes in 1 sec; and with
   repTarget = 6 page updates every 6 sec with a 100ms delay between events.
   No guru error evident (so far, running 2 days)

   Version 2.06 - added spiffs and sl2data from SolarLoggerDev2
   to log 6 min data to char dayFileName [16] = "/dayData.log";
   also enabled ftp server: all working with no guru errors

   *** GURU ERRORS NOW SORTED  ***

   Version 2.10: sl2data checkperiod & newday amended to save 6 min data to dayfile
   spiffs: add gatherday, appendint, deletefile

   Version 2.11 changed interrupt to rising edge as int pin is pulled DOWN when light on. Restored wifi event handling to manage wifi led.

   Version 2.12

   v 2.13 changed from "events.send(String(pWatts).c_str(), "pwatts", millis());" to
   "static char mypWatts[ 12] ; snprintf(mypWatts, 8, "%i",pWatts); events.send( mypWatts , "pwatts", millis());"
   still working fine.  Ready for upload to working unit in cupboard.

   V2.14, 2.15 - changed webevent data buffer in state machine to a single global; 
            incresed precision in float values to 3dp & reduced in processor
            WORKING WELL 
   V2.16: minor change: use int period=6; in sl2.h to set update interval in sl2data:checkPeriod();
   tested for count rate to be higher than required 
   (3.6 kW = 3600 pulses per hour = 60 per min = 1 per sec so we need to show more than 360 * 10 = 3600 cwatts)
   on test recorded over 1000 pulses in 6 min = 10kW;
   V2.17: on 29 Aug displayed values on webpage were incorrect BUT recorded values in day file seem OK.
      sl2.h - pCount AND cWatts are changed in the ISR; now marked as VOLATILE
   V2.18: try to set meaningful hostname
   V2.19: 01 Nov added setenv to correct to GMT & BST; not working
   V2.30: change to use LittleFS

*/

#include <WiFi.h>
#include <AsyncTCP.h>
#include <ESPAsyncWebServer.h>
#include <ESP8266FtpServer.h>

//#include <SPIFFS.h>
#include <LITTLEFS.h> //2.30
#define SPIFFS LITTLEFS

#include <time.h>  
//The use of angle brackets (<>) causes the compiler to search the default include directory. Should be used for library files.
//Double quotes ("") causes it to search the current working directory and, if that search fails, the default include directory.


#include "sl2.h"  //local header file

// Create AsyncWebServer object on port 80
AsyncWebServer server(80);
String hostname = "ESP32-SolarLogger-Test"; //attempt to give appropriate hostname
/*
 * Hostnames are composed of a sequence of labels concatenated with dots. 
 * For example, "en.wikipedia.org" is a hostname. 
 * Each label must be from 1 to 63 characters long.[2] 
 * The entire hostname, including the delimiting dots, has a maximum of 253 ASCII characters.[3]
 * The Internet standards (Requests for Comments) for protocols specify that labels may contain only the ASCII letters a through z (
 * in a case-insensitive manner), the digits 0 through 9, and the hyphen-minus character ('-').
 * WiFiGotIP() now reports hostname
 */

// Create an Event Source on /events
AsyncEventSource events("/events");

FtpServer ftpSrv;   //set #define FTP_DEBUG in ESP8266FtpServer.h to see ftp verbose on serial

void setup() {
  Serial.begin(115200);
  pinMode(LED_WiFi, OUTPUT);
  pinMode(LED_Pulse, OUTPUT);
  pinMode(intPin, INPUT_PULLUP);
  
  digitalWrite(LED_WiFi, 0); //ensure blue LED off

  // added in V2.11:  delete old wifi config
  WiFi.disconnect(true);
  //added in V2.18 to set a hostname
  WiFi.config(INADDR_NONE, INADDR_NONE, INADDR_NONE, INADDR_NONE);
  WiFi.setHostname(hostname.c_str()); //define hostname
  delay(200);

  //separate heading from reset diagnostic messages
  Serial.println("");
  Serial.println("Starting Solar Logger");

  //added in V2.11: monitor Wifi connection and handle LED - nb these dont work as is on 8266
  WiFi.onEvent(WiFiStationConnected, SYSTEM_EVENT_STA_CONNECTED);
  WiFi.onEvent(WiFiGotIP, SYSTEM_EVENT_STA_GOT_IP);
  WiFi.onEvent(WiFiStationDisconnected, SYSTEM_EVENT_STA_DISCONNECTED);

  initWiFi();
  delay(200);
  //  FTP Setup, ensure SPIFFS is started before ftp;
  if (SPIFFS.begin(true)) {
    Serial.println("File system opened!");
    ftpSrv.begin("esp8266", "esp8266");  //ftp username, password. set ports in ESP8266FtpServer.h (default 21, 50009 for PASV)
  }
  //configTzTime("GMT0BST,M3.5.0/1,M10.5.0", "pool.ntp.org"); //seems to work correctly across change from BST to GMT
  configTime(gmtOffset_sec, daylightOffset_sec, ntpServer);   // Init and get the time - to stay on GMT set daylightOffset_sec = 0
  //setenv("TZ","GMTGMT-1,M3.4.0/01,M10.4.0/02",1);  //
  /*
   * GMTGMT-1, M3.4.0/1,M10.4.0/2  
   * GMTGMT-1 is the name of the time zone  
   * GMT-1 is the abbreviation used when DST is off  1 hour is the time difference from GMT  
   * GMT is the abbreviation used when DST is on  ,
   * M3 is the third month  .4 is the forth occurrence of the day in the month when DST starts  .0 is Sunday  /1 is the time in hours when DST starts ,
   * M10 is the tenth month  .4 is the forth occurrence of the day in the month when DST ends  .0 is Sunday  /2 is the time in hours when DST ends
   */
  delay(100); //allow to get NTP time
  while (!getLocalTime(&timeinfo)) {
    delay(100);
  }
  currMonth = 1 + timeinfo.tm_mon; //tm_mon gives months since jan so eg April is 03 hence add 1
  currDay = timeinfo.tm_mday; //set current day
  Serial.printf("Setup: -  success getting time. currMonth is %i: currDay is %i \n", currMonth, currDay);
  Serial.println(&timeinfo, "%H:%M:%S");  //eg 07:04:45
  //currDay = 0; //testing - force new day scenario
  //currMonth = 0; //testing - force new month scenario
  newMonth();  //if no file exists as yet for month data create one; but if there is one dont overwrite it.

  // Handle Web Server
  server.on("/", HTTP_GET, [](AsyncWebServerRequest * request) {
    request->send_P(200, "text/html", index_html, processor);
  });

  // Handle Web Server Events
  events.onConnect([](AsyncEventSourceClient * client) {
    if (client->lastId()) {
      Serial.printf("Client reconnected! Last message ID that it got is: %u\n", client->lastId());
    }
    // send event with message "hello!", id current millis and set reconnect delay to 10 second
    client->send("hello!", NULL, millis(), 10000);
  });
  server.addHandler(&events);
  server.begin();

  zeroStats();
  configInterrupt();
  Serial.println("Setup complete \n");
}

void loop() {
  pulseState = digitalRead(intPin); //reflect input pulse to LED
  digitalWrite(LED_Pulse, (pulseState + 1) % 2); //because input active low
  ftpSrv.handleFTP();        //make sure in loop you call handleFTP()!!

  // 10 states so web update cycle completes every 100msec * 10 states * reptarget (=6) = 6 seconds
  if (millis() - previousMillis >= timerDelay) { // run the state machine once every 100msec
    previousMillis = millis();

    switch (state) {

      case 0: {
          if (!getLocalTime(&timeinfo)) {
            Serial.println("Failed to obtain time");
            return;
          }
          state = 1;
        }
        break;

      case 1: {
          checkPeriod(); //if 6 min elapsed, record current, previous, max values of watts

          if (timeinfo.tm_mday != currDay) {
            //start of new day
            newDay(); //save day data to file
            currDay = timeinfo.tm_mday; //update current day
          }
          if (timeinfo.tm_mon != currMonth) {
            //start of new month
            newMonth();
          }
          buildDate(); //only needed for web page date & time display
          state = 2;
        }
        break;

      // send web event updates

      case 2: {
          //Serial.printf("repcount is %i \n", repCount);
          if (repCount >= repTarget) {
            events.send("ping", NULL, millis());
          }
          state = 3;
        }
        break;

      case 3: {
          if (repCount >= repTarget) {
            myDate.toCharArray( myDateBuf, 20 ) ;  // myDate is already a String
            events.send( myDateBuf , "time", millis());
          }
          state = 4;
        }
        break;

      case 4: {
          if (repCount >= repTarget) {
            snprintf(webpageDataBuffer, 10, "%8.3f", dTotalkWh);  //buffer, max bytes, format specifier (%[flags][width][.precision][length]specifier), value
            events.send( webpageDataBuffer , "dtotal", millis());
          }
          state = 5;
        }
        break;

      case 5: {
          if (repCount >= repTarget) {
            snprintf(webpageDataBuffer, 10, "%8.3f", yTotalkWh);
            events.send( webpageDataBuffer , "ytotal", millis());
          }
          state = 6;
        }
        break;

      case 6: {
          if (repCount >= repTarget) {
            snprintf(webpageDataBuffer, 8, "%i", pWatts);
            events.send( webpageDataBuffer , "pwatts", millis());
          }
          state = 7;
        }
        break;

      case 7: {
          if (repCount >= repTarget) {
            snprintf(webpageDataBuffer, 8, "%i", mWatts);
            events.send( webpageDataBuffer , "mwatts", millis());
          }
          state = 8;
        }
        break;

      case 8: {
          if (repCount >= repTarget) {
            snprintf(webpageDataBuffer, 8, "%i", cWatts);
            events.send( webpageDataBuffer , "cwatts", millis());
            repCount = 0; //reset rep count
          }
          else repCount++;
          state = 9;
        }
        break;

      case 9: {
          state = 0;
        }
        break;
    }  //switch
  }//if its time to run again
}
